import sys
import xmlrpclib
import pkg_resources
from termcolor import colored
from nova.core.exc import NovaError


def check_latest_version():
    current = pkg_resources.require("gilt-nova")[0].version
    pypi = xmlrpclib.ServerProxy('http://pypi.python.org/pypi')
    available = pypi.package_releases('gilt-nova')
    major_available = available[0].split('.')[0]
    major_current = current.split('.')[0]
    if available[0] != current:
        print(colored("The latest version of nova is '%s', please upgrade!" % available[0], color='yellow'))

    if major_available != major_current:
        raise NovaError('There has been a breaking change, please upgrade before continuing!')


def get_git_revision():
    import subprocess
    return subprocess.check_output(['git', 'describe', '--tags', '--dirty', '--always']).strip('v').strip()


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    :param question: is a string that is presented to the user.
    :param default: is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = raw_input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write(colored("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n", color='cyan'))



import json
from collections import OrderedDict

import pyaml
from cfn_pyplates.core import CloudFormationTemplate
from termcolor import colored

from nova.core import query_yes_no
from nova.core.exc import NovaError
from nova.core.spec.environment import Environment
from nova.core.spec.service_log_mapping import ServiceLogMapping


class Service(object):

    def __init__(self, name, team_name, port, healthcheck_url, logs, environments):
        self.name = name
        self.team_name = team_name
        self.port = port
        self.healthcheck_url = healthcheck_url
        self.logs = logs
        self.environments = environments

    def yaml(self):
        logs_list = None
        if self.logs is not None:
            logs_list = map(lambda lm: lm.yaml(), self.logs)

        data = OrderedDict([
            ('service_name', self.name),
            ('team_name', self.team_name),
            ('port', self.port),
            ('healthcheck_url', self.healthcheck_url),
            ('logs', logs_list),
            ('environments', map(lambda e: e.yaml(), self.environments))
        ])
        return pyaml.dump(OrderedDict((k,v) for k,v in data.iteritems() if v is not None))

    def get_environment(self, environment_name):
        environment = next((item for item in self.environments if item.name == environment_name), None)
        if environment is not None:
            return environment
        else:
            raise NovaError("Environment '%s' was not found in the configuration" % environment_name)

    def set_code_deploy_app(self, environment_name, code_deploy_app_id):
        self.get_environment(environment_name).deployment_application_id = code_deploy_app_id

    def to_cfn_template(self, environment, template_bucket, aws_profile):
        description = "%s %s %s stack" % (self.team_name, self.name, environment.name)
        cfn = CloudFormationTemplate(description=description)
        environment.to_cfn_template(self, template_bucket, cfn, aws_profile)
        json_cfn = cfn.json

        # if query_yes_no("Output Cloudformation to nova-cloudformation.json?"):
        #     with open('nova-cloudformation.json', 'w') as out_json:
        #         out_json.write(json.dumps(json.loads(json_cfn), indent=2))

        return json_cfn

    @staticmethod
    def load(values):
        logs_list = values.get("logs")
        if logs_list is not None:
            logs_list = map(lambda lm: ServiceLogMapping.load(lm), logs_list)

        return Service(
            values.get("service_name"),
            values.get("team_name"),
            values.get("port"),
            values.get("healthcheck_url"),
            logs_list,
            map(lambda e: Environment.load(e), values.get("environments"))
        )



import base64
import pickle
from collections import OrderedDict

import boto3
from cfn_pyplates.core import Resource, Output
from cfn_pyplates.functions import ref, get_att


def stack_param_known(k):
    return k.startswith('stack_') or k.startswith('deployment_')


class Stack(object):
    base_template_url='https://s3.amazonaws.com/%s/%s/%s.json'
    base_custom_template_url='https://s3.amazonaws.com/%s/%s/%s_%s'

    def __init__(self, name, stack_type, custom_template, template, template_version, deploy_config, deployment_options,
                 deployment_volumes, deployment_variables, deployment_arguments, deployment_group, extra_parameters):
        self.name = name
        self.stack_type = stack_type
        self.custom_template = custom_template
        self.template = template
        self.template_version = template_version
        self.deploy_config = deploy_config
        self.deployment_options = deployment_options
        self.deployment_volumes = deployment_volumes
        self.deployment_variables = deployment_variables
        self.deployment_arguments = deployment_arguments
        self.deployment_group = deployment_group
        self.extra_parameters = extra_parameters

    def yaml(self):
        data = OrderedDict([
            ('stack_name', self.name),
            ('stack_type', self.stack_type),
            ('stack_custom_template', self.custom_template),
            ('stack_template', self.template),
            ('stack_template_version', self.template_version),
            ('stack_deploy_config', self.deploy_config),
            ('deployment_options', self.deployment_options),
            ('deployment_volumes', self.deployment_volumes),
            ('deployment_variables', self.deployment_variables),
            ('deployment_arguments', self.deployment_arguments),
            ('deployment_group', self.deployment_group)
        ])
        data.update(self.extra_parameters)
        return OrderedDict((k,v) for k,v in data.iteritems() if v is not None)

    def to_cfn_template(self, service, environment, template_bucket, codedeploy_app_name, cft, aws_profile):
        self.__add_stack(cft, service, environment, template_bucket, aws_profile)
        self.__add_deployment_group(cft, codedeploy_app_name, environment.deploy_arn)

    def __add_stack(self, cft, service, environment, template_bucket, aws_profile):
        other_params = self.extra_parameters.copy()

        if other_params.has_key('DNS') and other_params.get('HostedZoneName') is None:
            record = other_params.get('DNS')
            awsprofile = aws_profile or environment.aws_profile
            session = boto3.session.Session(profile_name=awsprofile, region_name=environment.aws_region)
            route53 = session.client('route53')
            hostedzones = map(lambda z: z.get('Name'), route53.list_hosted_zones_by_name().get('HostedZones'))
            hostedzone = next((z for z in hostedzones if (z in record) or (z[:len(z)-1] in record)), None)
            other_params['HostedZoneName'] = hostedzone

        other_params.update([
            ('StackType', self.stack_type),
            ('Port', service.port),
            ('ApplicationName', service.name),
            ('TeamName', service.team_name),
            ('HealthcheckUrl', service.healthcheck_url),
            ('LogsList', self.serialize_logs_settings(service.logs)),
            ('DockerDeploymentOptions', self.pickle_encode(self.deployment_options)),
            ('DockerDeploymentVariables', self.pickle_encode(self.deployment_variables)),
            ('DockerDeploymentVolumes', self.pickle_encode(self.deployment_volumes)),
            ('DockerDeploymentArguments', self.pickle_encode(self.deployment_arguments))
        ])
        parameters = {
            'TemplateURL': self.template_url_to_use(service, template_bucket),
            'TimeoutInMinutes': 60,
            'Parameters': other_params
        }
        cft.resources.add(Resource(self.name.title(), 'AWS::CloudFormation::Stack', parameters))

    def __add_deployment_group(self, cft, codedeploy_app_name, deploy_arn):
        parameters = {
            'ApplicationName': ref(codedeploy_app_name),
            'AutoScalingGroups': [get_att(self.name.title(), 'Outputs.AutoScalingGroup')],
            'DeploymentConfigName': 'CodeDeployDefault.%s' % self.deploy_config,
            'ServiceRoleArn': deploy_arn
        }
        deployment_group_name = self.name.title() + 'DeploymentGroup'
        cft.resources.add(Resource(deployment_group_name, 'AWS::CodeDeploy::DeploymentGroup', parameters))
        cft.outputs.add(Output(deployment_group_name, ref(deployment_group_name)))

    def get_template_used(self, service, template_bucket):
        if self.custom_template is not None:
            return {'filename': self.custom_template, 's3_key': '%s/%s_%s' % (service.name, self.name, self.custom_template)}
        else:
            return {'name': self.template, 'version': self.template_version}

    def template_url_to_use(self, service, template_bucket):
        if self.custom_template is not None:
            return self.base_custom_template_url % (template_bucket, service.name, self.name, self.custom_template)
        else:
            return self.base_template_url % (template_bucket, self.template_version, self.template)

    @staticmethod
    def load(values):
        extra_parameters = dict([(k, v) for (k, v) in values.iteritems() if not stack_param_known(k)])
        return Stack(
            values.get("stack_name"),
            values.get('stack_type'),
            values.get('stack_custom_template'),
            values.get("stack_template"),
            values.get("stack_template_version"),
            values.get("stack_deploy_config"),
            values.get("deployment_options"),
            values.get("deployment_volumes"),
            values.get("deployment_variables"),
            values.get("deployment_arguments"),
            values.get("deployment_group"),
            extra_parameters
        )

    @staticmethod
    def stack_param_known(k):
        return k.startswith('stack_') or k.startswith('deployment_')

    def serialize_logs_settings(self, logs):
        if logs is not None:
            log_list = map(lambda l: l.yaml(), logs)
            return self.pickle_encode(log_list)
        else:
            return base64.b64encode(pickle.dumps([]))

    @staticmethod
    def pickle_encode(to_pickle):
        if to_pickle is not None:
            return base64.b64encode(pickle.dumps(to_pickle))
        else:
            return base64.b64encode(pickle.dumps([]))



"""nova base controller."""

from cement.core.controller import CementBaseController, expose
from nova.core import check_latest_version
import pkg_resources


VERSION = pkg_resources.require("gilt-nova")[0].version

BANNER = """
NOVA Service Tools v%s
Copyright (c) 2016 Gilt Groupe
""" % VERSION


class NovaBaseController(CementBaseController):
    class Meta:
        label = 'base'
        description = 'NOVA service tools'
        arguments = [
            (['-v', '--version'], dict(action='version', version=BANNER)),
        ]

    @expose(hide=True)
    def default(self):
        check_latest_version()
        self.app.args.print_help()



import os
import yaml

from cerberus import Validator

from nova.core import spec
from nova.core.exc import NovaError
from nova.core.spec import yaml_include
from nova.core.spec.service import Service


class NovaServiceLoader:

    def __init__(self, environment=None, nova_descriptor_file=None):
        if nova_descriptor_file is None:
            nova_descriptor_file = "nova.yml"

        self.environment = environment
        self.templates_used = dict()
        yaml.add_constructor("!include", yaml_include)

        with open(os.path.join(spec.__path__[0], 'nova_service_schema.yml'), 'r') as schemaYaml:
            schema = yaml.load(schemaYaml)

        v = Validator(schema)
        try:
            with open(nova_descriptor_file, 'r') as novaYaml:
                self.service_spec = yaml.safe_load(novaYaml)

            # Validate loaded dictionary
            valid = v.validate(self.service_spec)
            if not valid:
                raise NovaError("Invalid nova service descriptor file '%s': %s" % (nova_descriptor_file, v.errors))
            else:
                self.service = Service.load(self.service_spec)
        except IOError:
            raise NovaError("No nova service descriptor found at '%s'" % nova_descriptor_file)

    def set_code_deploy_app(self, environment_name, code_deploy_app_id):
        self.service.set_code_deploy_app(environment_name, code_deploy_app_id)
